import numpy as np
from .utils import _row_f, _check_dissimilarity_matrix


def upper_bound_samples(D: np.ndarray, kappa: int = 1) -> np.ndarray:
    """
    Compute a sharp upper bound of the Silhouette coefficient for each data point.

    Parameters
    ----------
    D: np.ndarray
        Square matrix of pairwise distances (or dissimilarities) (shape: [n_samples, n_samples]).
    kappa: int
        Lower limit of cluster size (default is 1).

    Returns
    -------
    np.ndarray
        A 1D array where the i:th element is a sharp upper bound of the Silhouette coefficient s(i).

    References
    ----------
    .. [1] Silhouette (clustering). Wikipedia. https://en.wikipedia.org/wiki/Silhouette_(clustering)
    """

    _check_dissimilarity_matrix(D=D)

    # Remove diagonal from distance matrix and then sort
    D_hat = np.sort(D[~np.eye(D.shape[0], dtype=bool)].reshape(D.shape[0], -1))

    n = D_hat.shape[0]
    if n < 4:
        raise ValueError("Matrix must be at least of size 4x4.")

    if kappa < 1 or kappa > n - 1:
        raise ValueError("The parameter kappa is out of range.")

    # Compute bounds
    bounds = np.apply_along_axis(
        lambda row: _row_f(row, kappa=kappa, n=n), axis=1, arr=D_hat
    )

    return bounds


def upper_bound(D: np.ndarray, kappa: int = 1) -> float:
    """
    Compute an upper bound of the Average Silhouette Width (ASW). The upper bound ranges from 0 to 1.

    Parameters
    ----------
    D: np.ndarray
        Square matrix of pairwise distances (or dissimilarities) (shape: [n_samples, n_samples]).
    kappa: int
        Lower limit of cluster size (default is 1).

    Returns
    -------
    float
        An upper bound of the ASW.

    Notes
    -----
    We emphasize that the upper bound is not guaranteed to be close to the true global ASW-maximum.
    Comparison with outputs from suitable clustering algorithms is advised.

    References
    ----------
    .. [1] Silhouette (clustering). Wikipedia. https://en.wikipedia.org/wiki/Silhouette_(clustering)
    """

    point_bounds = upper_bound_samples(D=D, kappa=kappa)

    return np.mean(point_bounds)



from silhouette_upper_bound import upper_bound_samples
from sklearn.datasets import make_blobs
from sklearn.metrics import pairwise_distances
from utils import kmeans_optimized, hierarchical_optimized, Counter
import numpy as np


def table(rows):

    headers = [
        "Dataset",
        "Metric",
        "Hierarchical weighted",
        "Hierarchical single",
        "KMeans",
        "Upper bound",
        "Min",
        "Max",
    ]

    # Format header
    header_line = "| " + " | ".join(headers) + " |"
    separator = "| " + " | ".join(["---"] * len(headers)) + " |"

    print(header_line)
    print(separator)

    # Format rows
    for row in rows:

        print(" & ".join(f"${str(cell)}$" for cell in row) + " \\\ ")


def table_row(params):

    n_samples, n_features, centers, cluster_std = params
    # Generate synthetic data
    X, _ = make_blobs(
        n_samples=n_samples,
        n_features=n_features,
        centers=centers,
        cluster_std=cluster_std,
        random_state=0,
    )
    D = pairwise_distances(X)

    # Compute upper bound
    ubs = upper_bound_samples(D)
    ub = np.mean(ubs)
    ubs_min = np.min(ubs)
    ubs_max = np.max(ubs)

    # Kmeans
    kmeans_dict = kmeans_optimized(data=X, n_init=10)

    # Single
    single_dict = hierarchical_optimized(data=X, metric="euclidean", method="single")

    # Weigthed
    weighted_dict = hierarchical_optimized(
        data=X, metric="euclidean", method="weighted"
    )

    kmeans_str = f"${kmeans_dict['best_score']:.3f}$ ({len(Counter(kmeans_dict['best_labels']))})"
    weighted_str = f"${weighted_dict['best_score']:.3f}$ ({len(Counter(weighted_dict['best_labels']))})"
    single_str = f"${single_dict['best_score']:.3f}$ ({len(Counter(single_dict['best_labels']))})"

    return (
        "-".join(str(x) for x in params),
        "Euclidean",
        weighted_str,
        single_str,
        kmeans_str,
        ub,
        ubs_min,
        ubs_max,
    )


def table(caseparams: list):
    """
    Print table in terminal.
    """

    headers = [
        "Dataset",
        "Metric",
        "Hierarchical weighted",
        "Hierarchical single",
        "KMeans",
        "UB(D)",
        "minUB(D)",
        "maxUB(D)",
    ]

    lines = []

    # Format header
    header_line = "| " + " | ".join(headers) + " |"
    lines.append(header_line)
    separator = "| " + " | ".join(["---"] * len(headers)) + " |"
    lines.append(separator)

    for params in caseparams:
        row = table_row(params=params)

        lines.append(
            " & ".join(
                f"${cell:.3f}$" if type(cell) is not str else f"{cell}" for cell in row
            )
            + " \\\ "
        )

    # Print table to terminal
    print("\nTABLE\n")
    for line in lines:
        print(line)


if __name__ == "__main__":

    # n_samples, n_features, n_centers, cluster_std
    case1params = (400, 64, 5, 6)
    case2params = (400, 64, 2, 2)
    case3params = (400, 128, 7, 3)
    case4params = (1000, 161, 2, 13)
    case5params = (1000, 300, 5, 2)
    case6params = (10000, 32, 20, 2)
    case7params = (10000, 1024, 20, 4)

    table(
        [
            case1params,
            case2params,
            case3params,
            case4params,
            case5params,
            case6params,
            case7params,
        ]
    )



"""
In this example, we cluster a synthetic dataset using KMeans.
We then plot the Silhouette score of each sample and compare it to its upper bound.
"""

import matplotlib.pyplot as plt
import numpy as np
from silhouette_upper_bound import upper_bound, upper_bound_samples
from sklearn.cluster import KMeans
from sklearn.datasets import make_blobs
from sklearn.metrics import silhouette_score, silhouette_samples, pairwise_distances


def main():
    # Generate synthetic data
    X, _ = make_blobs(
        n_samples=500, n_features=16, centers=3, cluster_std=2, random_state=42
    )
    D = pairwise_distances(X)

    # Cluster with KMeans
    kmeans = KMeans(n_clusters=3, random_state=42, n_init="auto")
    labels = kmeans.fit_predict(X)

    # Overall silhouette score
    score = silhouette_score(X, labels)

    # Upper bound
    ub = upper_bound(D)

    # Upper bound for each sample
    ub_samples = upper_bound_samples(D)

    # Compute silhouette values for each sample
    sample_silhouette_values = silhouette_samples(X, labels)

    fig, ax = plt.subplots(figsize=(8, 5))

    # Silhouette plot setup
    ax.set_xlim([-0.1, 1])
    ax.set_ylim([0, len(X) + (3 + 1) * 10])  # 3 clusters + padding

    y_lower = 10
    for i in range(3):

        indices = np.where(labels == i)[0]
        cluster_silhouettes = sample_silhouette_values[indices]
        cluster_ub_values = ub_samples[indices]

        # Get sorted order of silhouette values
        sorted_order = np.argsort(cluster_silhouettes)

        sorted_silhouettes = cluster_silhouettes[sorted_order]
        sorted_ub_values = cluster_ub_values[sorted_order]

        size_cluster_i = sorted_silhouettes.shape[0]
        y_upper = y_lower + size_cluster_i

        color = plt.cm.viridis(float(i) / 3)

        # Cluster Silhouette scores
        ax.fill_betweenx(
            np.arange(y_lower, y_upper),
            0,
            sorted_silhouettes,
            facecolor=color,
            edgecolor=color,
            alpha=0.7,
        )

        # Cluster Silhouette bounds
        ax.fill_betweenx(
            np.arange(y_lower, y_upper),
            0,
            sorted_ub_values,
            facecolor=color,
            edgecolor=color,
            alpha=0.2,
        )

        # Label cluster number
        ax.text(-0.05, y_lower + 0.5 * size_cluster_i, str(i))

        y_lower = y_upper + 10  # 10 for spacing between clusters

    ax.set_title("Silhouette plot for clusters")
    ax.set_xlabel("Silhouette coefficient values and upper bounds (shaded)")
    ax.set_ylabel("Cluster label")
    ax.axvline(x=score, color="red", linestyle="--", label="Average silhouette score")
    ax.axvline(x=ub, color="black", linestyle="--", label="Upper bound")
    ax.set_yticks([])  # Clear y-axis labels for clarity
    ax.legend()

    plt.tight_layout()
    plt.show()


if __name__ == "__main__":
    main()



import utils


def table_row(dataset: str, metric: str):

    print(f"\nDistance metric: {metric}")

    if dataset == "conference_papers":
        data = utils.get_data(dataset=dataset, transpose=True)
    else:
        data = utils.get_data(dataset=dataset)

    n = data.shape[0]

    ub_dict = utils.get_upper_bound(data=data, metric=metric)

    # Weighted
    if n > 1000:
        weighted_dict = utils.hierarchical_optimized(
            data=data, metric=metric, method="weighted", t_range=range(2, 51)
        )
    else:
        weighted_dict = utils.hierarchical_optimized(
            data=data, metric=metric, method="weighted", t_range=range(2, n)
        )

    # Single
    if n > 1000:
        single_dict = utils.hierarchical_optimized(
            data=data, metric=metric, method="single", t_range=range(2, 51)
        )
    else:
        single_dict = utils.hierarchical_optimized(
            data=data, metric=metric, method="single", t_range=range(2, n)
        )

    # Kmeans
    if metric == "euclidean":
        kmeans_dict = utils.kmeans_optimized(data=data, k_range=range(2, 51))
        kmeans_str = f"${kmeans_dict['best_score']:.3f}$ ({len(utils.Counter(kmeans_dict['best_labels']))})"
    else:
        kmeans_dict = {"best_score": "N/A"}
        kmeans_str = "N/A"

    weighted_str = f"${weighted_dict['best_score']:.3f}$ ({len(utils.Counter(weighted_dict['best_labels']))})"
    single_str = f"${single_dict['best_score']:.3f}$ ({len(utils.Counter(single_dict['best_labels']))})"

    return [
        dataset,
        metric,
        weighted_str,
        single_str,
        kmeans_str,
        ub_dict["ub"],
        ub_dict["min"],
        ub_dict["max"],
    ]


def table(dataset_metric: list):
    """
    Print table in terminal.
    """

    headers = [
        "Dataset",
        "Metric",
        "Hierarchical weighted",
        "Hierarchical single",
        "KMeans",
        "UB(D)",
        "minUB(D)",
        "maxUB(D)",
    ]

    lines = []

    # Format header
    header_line = "| " + " | ".join(headers) + " |"
    lines.append(header_line)
    separator = "| " + " | ".join(["---"] * len(headers)) + " |"
    lines.append(separator)

    for dataset, metric in dataset_metric:
        row = table_row(dataset=dataset, metric=metric)

        lines.append(
            " & ".join(
                f"${cell:.3f}$" if type(cell) is not str else f"{cell}" for cell in row
            )
            + " \\\ "
        )

    # Print table to terminal
    print("\nTABLE\n")
    for line in lines:
        print(line)


if __name__ == "__main__":

    dataset_metric = [
        ("rna", "correlation"),
        ("religious_texts", "cosine"),
        ("conference_papers", "cosine"),
        ("religious_texts", "euclidean"),
        ("ceramic", "euclidean"),
        ("conference_papers", "euclidean"),
        ("rna", "euclidean"),
        ("religious_texts", "jaccard"),
        ("conference_papers", "jaccard"),
    ]

    table(dataset_metric=dataset_metric)



# sequence A000110 in the OEIS
bell_numbers = [
    1,
    1,
    2,
    5,
    15,
    52,
    203,
    877,
    4140,
    21147,
    115975,
    678570,
    4213597,
    27644437,
    190899322,
    1382958545,
    10480142147,
    82864869804,
    682076806159,
    5832742205057,
    51724158235372,
    474869816156751,
    4506715738447323,
    44152005855084346,
    445958869294805289,
    4638590332229999353,
    49631246523618756274,
]


import matplotlib.pyplot as plt
import seaborn as sns
import pandas as pd

# Create DataFrame
df = pd.DataFrame({"n": range(len(bell_numbers)), "B_n": bell_numbers})

# Set seaborn theme
sns.set_theme(style="whitegrid", context="talk", font_scale=1.0)

# Initialize figure
fig, ax = plt.subplots(figsize=(10, 6))

# Plot with seaborn
sns.scatterplot(
    data=df,
    x="n",
    y="B_n",
    ax=ax,
    color=sns.color_palette("deep")[0],
    s=100,
    edgecolor="black",
    linewidth=0.5,
)

# Log scale
ax.set_yscale("log")

ax.tick_params(axis="y", which="major")

sns.despine(ax=ax)

ax.set_xlabel(r"$n$", fontsize=14, labelpad=10)
ax.set_ylabel(r"$B_n$", fontsize=14, labelpad=10)

plt.tight_layout()
plt.savefig("bell_numbers.pdf")



"""
In this example, we show how to use the standard upper bound on a synthetic dataset.
"""

from silhouette_upper_bound import upper_bound
from sklearn.cluster import KMeans
from sklearn.datasets import make_blobs
from sklearn.metrics import silhouette_score, pairwise_distances


def main():
    # Generate synthetic data
    X, _ = make_blobs(n_samples=3000, n_features=32, centers=7, random_state=42)
    D = pairwise_distances(X)

    # Cluster with KMeans
    kmeans = KMeans(n_clusters=7, random_state=42, n_init="auto")
    labels = kmeans.fit_predict(X)

    # Overall silhouette score
    score = silhouette_score(X, labels)
    print(f"KMeans Silhouette score: {score:.3f}")

    # Upper bound
    print(f"Silhouette upper bound: {upper_bound(D):.3f}")


if __name__ == "__main__":
    main()



import numpy as np
from silhouette_upper_bound import upper_bound_samples
from sklearn.datasets import make_blobs
from sklearn.metrics import pairwise_distances
from utils import kmeans_optimized, hierarchical_optimized
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator
import pandas as pd
import seaborn as sns


def graph(params):

    # Parameters
    n_samples, n_features, centers, cluster_std = params

    # Generate synthetic data
    X, _ = make_blobs(
        n_samples=n_samples,
        n_features=n_features,
        centers=centers,
        cluster_std=cluster_std,
        random_state=0,
    )
    D = pairwise_distances(X)
    ref = np.mean(upper_bound_samples(D=D))

    ref_80 = np.mean(upper_bound_samples(D=D, kappa=80))

    # Storage
    k_list = []
    silh_list = []
    single_list = []
    weighted_list = []

    for k in range(2, 21):

        # Kmeans
        kmeans_scores = kmeans_optimized(data=X, k_range=range(k, k + 1), n_init=10)[
            "best_scores"
        ]

        # Single
        single_scores = hierarchical_optimized(
            data=X, metric="euclidean", method="single", t_range=range(k, k + 1)
        )["best_scores"]

        # Weighted
        weighted_scores = hierarchical_optimized(
            data=X, metric="euclidean", method="weighted", t_range=range(k, k + 1)
        )["best_scores"]

        k_list.append(k)
        silh_list.append(np.mean(kmeans_scores))
        single_list.append(np.mean(single_scores))
        weighted_list.append(np.mean(weighted_scores))

    # Put data into a tidy DataFrame for seaborn
    df = pd.DataFrame(
        {
            "K": k_list,
            "KMeans Silhouette": silh_list,
            "Weighted-Linkage Silhouette": weighted_list,
            "Single-Linkage Silhouette": single_list,
        }
    )

    # Melt the DataFrame for seaborn
    df_melted = df.melt(
        id_vars="K",
        value_vars=[
            "KMeans Silhouette",
            "Weighted-Linkage Silhouette",
            "Single-Linkage Silhouette",
        ],
        var_name="Method",
        value_name=" ",
    )

    # Plot
    sns.set(style="whitegrid", context="talk")

    plt.figure(figsize=(10, 6))
    ax = sns.lineplot(
        data=df_melted,
        x="K",
        y=" ",
        hue="Method",
        style="Method",
        markers=True,
        dashes=False,
        linewidth=2.5,
    )

    # Reference lines
    print(f"Upper bound = {ref}")
    print(f"Upper bound (kappa 80) = {ref_80}")
    plt.axhline(
        y=ref, color="black", linestyle="--", linewidth=1.5, label=f"Upper bound"
    )
    plt.axhline(
        y=ref_80,
        color="red",
        linestyle="--",
        linewidth=1.5,
        label=rf"Upper bound ($\kappa$={80})",
    )

    # Adjust axes
    ax.xaxis.set_major_locator(MaxNLocator(integer=True))
    plt.ylim(0, 0.4)

    # Titles and labels
    plt.xlabel("Number of clusters (K)", fontsize=14)
    plt.legend(fontsize=12, title_fontsize=13, loc="best")

    plt.tight_layout()
    plt.savefig("koptim.pdf")


if __name__ == "__main__":

    caseparams = (400, 64, 5, 6)
    graph(params=caseparams)



from silhouette_upper_bound import upper_bound_samples
from sklearn.cluster import KMeans
from scipy.cluster.hierarchy import linkage, fcluster
from scipy.spatial.distance import squareform
from sklearn.metrics import silhouette_samples, pairwise_distances
import numpy as np
import pandas as pd
from tqdm import tqdm
from collections import Counter


def get_data(dataset: str, transpose=False) -> np.ndarray:
    path = f"data/{dataset}/data.csv"

    print(f"==== Running dataset: {dataset} ====\n")

    df = pd.read_csv(path)

    data = df.select_dtypes(include="number")

    if transpose:
        data = data.transpose()

    print(f"Data shape: {data.shape}")

    data = data.to_numpy()

    # Removing zero-vectors
    non_zero_rows = ~np.all(data == 0, axis=1)

    data = data[non_zero_rows]

    print(f"Data shape (zeros removed): {data.shape}")

    return data


def _optim_iteration(data, cluster_labels, metric, best_solution):

    try:
        silh_samples = silhouette_samples(data, cluster_labels, metric=metric)
    except:
        silh_samples = np.zeros(data.shape[0])

    silh_score = np.mean(silh_samples)

    if silh_score > best_solution["best_score"]:

        best_solution["best_score"] = silh_score
        best_solution["best_scores"] = silh_samples
        best_solution["best_labels"] = cluster_labels

    return best_solution


def kmeans_optimized(data, k_range=range(2, 31), random_state=42, n_init="auto"):
    """

    Parameters
    ----------
        data: np.ndarray
            array of shape (n_samples, n_features)
        k_range: range
            specifies which values for K to use
    """

    best_solution = {
        "best_score": 0,  # ASW
        "best_scores": None,  # Silhouette samples
        "best_labels": None,  # Cluster labels
    }

    print(f"Kmeans optimization")

    for k in tqdm(k_range):

        kmeans = KMeans(n_clusters=k, random_state=random_state, n_init=n_init)

        cluster_labels = kmeans.fit_predict(data) + 1  # 1:indexed

        best_solution = _optim_iteration(
            data=data,
            cluster_labels=cluster_labels,
            metric="euclidean",
            best_solution=best_solution,
        )

    n_clusters = len(Counter(best_solution["best_labels"]))
    if n_clusters <= 6:
        clusters = {int(k): v for k, v in Counter(best_solution["best_labels"]).items()}
        print(
            f"best score: {best_solution['best_score']} | n clusters: {n_clusters} | clusters: {sorted(clusters.items())}"
        )
    else:
        print(f"best score: {best_solution['best_score']} | n clusters: {n_clusters}")

    return best_solution


def hierarchical_optimized(
    data: np.ndarray, metric: str, method="single", t_range=range(2, 31), TOL=1e-10
):
    """

    Parameters
    ----------
        D: np.ndarray
            distance matrix of shape (n_samples, n_features)
        metric: str
            distance metric, e.g. "euclidean" or "cosine"
        method: str
            specifies the linkage method
        t_range: range
            specifies which values for K to use

    """

    D = pairwise_distances(data, metric=metric)  # convert data to dissimilarity matrix

    assert np.linalg.norm(D - D.T, ord="fro") < TOL, f"Matrix X is not symmetric!"
    assert (
        np.abs(np.diag(D)).max() < TOL
    ), f"Diagonal entries of X are not close to zero!"

    vector_D = squareform(
        D, checks=False
    )  # convert dissimilarity matrix to vector-form distance vector

    best_solution = {
        "best_score": 0,  # ASW
        "best_scores": None,  # Silhouette samples
        "best_labels": None,  # Cluster labels
    }

    print(f"{method} optimization")
    for t in tqdm(t_range):

        cluster_labels = fcluster(
            linkage(vector_D, method=method), t=t, criterion="maxclust"
        )

        best_solution = _optim_iteration(
            data=D,
            cluster_labels=cluster_labels,
            metric="precomputed",
            best_solution=best_solution,
        )

    print(
        f"best score: {best_solution['best_score']} | n clusters: {len(Counter(best_solution['best_labels']))}"
    )

    return best_solution


def get_upper_bound(data: np.ndarray, metric: str) -> dict:

    D = pairwise_distances(data, metric=metric)  # convert data to dissimilarity matrix

    print(f"Computing upper bound")

    ubs = upper_bound_samples(D)

    ub = np.mean(ubs)
    ubs_min = np.min(ubs)
    ubs_max = np.max(ubs)

    print(f"UB: {ub}")

    return {"ub": ub, "min": ubs_min, "max": ubs_max, "samples": ubs}
